"""
API reference documentation for `behaviors` submodule.
"""

import egttools.behaviors.CRD as CRD
import egttools.behaviors.NormalForm as NormalForm
import egttools.behaviors.pgg_behaviors as pgg_behaviors

__all__ = ['CRD', 'NormalForm', 'pgg_behaviors']
