/*Tweaks to the Pydata default CSS */

/*No yellow background highlight when targeted by summary tables */
/*dt:target { background-color: #f8f8f8; border: 1px solid black, }*/
dt:target { background: transparent;}
/*More space between H1s and signatures in API reference*/
h1 { margin-bottom: 40px; }

/*No line underneath summary table headings (clashes with line above first member)*/
p.rubric { border-bottom: 0px; }