//
// Created by Elias Fernandez on 14/11/2022.
//
#pragma once
#ifndef EGTTOOLS_PYBIND11FILES_GAMES_HPP
#define EGTTOOLS_PYBIND11FILES_GAMES_HPP

#include <egttools/finite_populations/games/NormalFormGame.h>
#include <egttools/finite_populations/games/NormalFormNetworkGame.h>

#include <egttools/finite_populations/games/AbstractGame.hpp>
#include <egttools/finite_populations/games/AbstractNPlayerGame.hpp>
#include <egttools/finite_populations/games/AbstractSpatialGame.hpp>
#include <egttools/finite_populations/games/CRDGame.hpp>
#include <egttools/finite_populations/games/CRDGameTU.hpp>
#include <egttools/finite_populations/games/Matrix2PlayerGameHolder.hpp>
#include <egttools/finite_populations/games/MatrixNPlayerGameHolder.hpp>
#include <egttools/finite_populations/games/NPlayerStagHunt.hpp>
#include <egttools/finite_populations/games/OneShotCRD.hpp>
#include <egttools/finite_populations/games/OneShotCRDNetworkGame.hpp>

#include "egttools_common.hpp"
#include "python_stubs.hpp"

#endif//EGTTOOLS_PYBIND11FILES_GAMES_HPP
