//
// Created by Elias Fernandez on 14/11/2022.
//
#pragma once
#ifndef EGTTOOLS_PYBIND11FILES_DATASTRUCTURES_HPP
#define EGTTOOLS_PYBIND11FILES_DATASTRUCTURES_HPP

#include <egttools/Data.hpp>

#include "egttools_common.hpp"

#endif//EGTTOOLS_PYBIND11FILES_DATASTRUCTURES_HPP
