//
// Created by Elias Fernandez on 14/11/2022.
//

#ifndef EGTTOOLS_PYBIND11FILES_BEHAVIORS_HPP
#define EGTTOOLS_PYBIND11FILES_BEHAVIORS_HPP

#include <egttools/finite_populations/behaviors/CRDStrategies.h>

#include <egttools/finite_populations/behaviors/AbstractCRDStrategy.hpp>
#include <egttools/finite_populations/behaviors/AbstractNFGStrategy.hpp>
#include <egttools/finite_populations/behaviors/NFGStrategies.hpp>

#include "egttools_common.hpp"
#include "python_stubs.hpp"

#endif//EGTTOOLS_PYBIND11FILES_BEHAVIORS_HPP
