//
// Created by Elias Fernandez on 13/11/2022.
//

#pragma once
#ifndef EGTTOOLS_EGTTOOLS_COMMON_HPP
#define EGTTOOLS_EGTTOOLS_COMMON_HPP

#include <egttools/Types.h>
#include <pybind11/eigen.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <memory>
#include <stdexcept>
#include <utility>

namespace py = pybind11;

#endif//EGTTOOLS_EGTTOOLS_COMMON_HPP
